# -*- coding: utf-8 -*-
# PyonFX: An easy way to create KFX (Karaoke Effects) and complex typesetting using the ASS format (Advanced Substation Alpha).
# Copyright (C) 2019 Antonio Strippoli (CoffeeStraw/YellowFlash)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyonFX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

from __future__ import annotations
import re
import os
import sys
import time
import copy
import subprocess
from typing import List, Tuple, Union, Optional

from .font_utility import Font
from .convert import Convert


class Meta:
    """Meta object contains informations about the Ass.

    More info about each of them can be found on http://docs.aegisub.org/manual/Styles

    Attributes:
        wrap_style (int): Determines how line breaking is applied to the subtitle line
        scaled_border_and_shadow (bool): Determines if it has to be used script resolution (*True*) or video resolution (*False*) to scale border and shadow
        play_res_x (int): Video Width
        play_res_y (int): Video Height
        audio (str): Loaded audio path (absolute)
        video (str): Loaded video path (absolute)
    """

    wrap_style: int
    scaled_border_and_shadow: bool
    play_res_x: int
    play_res_y: int
    audio: str
    video: str

    def __repr__(self):
        return pretty_print(self)


class Style:
    """Style object contains a set of typographic formatting rules that is applied to dialogue lines.

    More info about styles can be found on http://docs.aegisub.org/3.2/ASS_Tags/.

    Attributes:
        fontname (str): Font name
        fontsize (float): Font size in points
        color1 (str): Primary color (fill)
        alpha1 (str): Transparency of color1
        color2 (str): Secondary color (secondary fill, for karaoke effect)
        alpha2 (str): Transparency of color2
        color3 (str): Outline (border) color
        alpha3 (str): Transparency of color3
        color4 (str): Shadow color
        alpha4 (str): Transparency of color4
        bold (bool): Font with bold
        italic (bool): Font with italic
        underline (bool): Font with underline
        strikeout (bool): Font with strikeout
        scale_x (float): Text stretching in the horizontal direction
        scale_y (float): Text stretching in the vertical direction
        spacing (float): Horizontal spacing between letters
        angle (float): Rotation of the text
        border_style (bool): *True* for opaque box, *False* for standard outline
        outline (float): Border thickness value
        shadow (float): How far downwards and to the right a shadow is drawn
        alignment (int): Alignment of the text
        margin_l (int): Distance from the left of the video frame
        margin_r (int): Distance from the right of the video frame
        margin_v (int): Distance from the bottom (or top if alignment >= 7) of the video frame
        encoding (int): Codepage used to map codepoints to glyphs
    """

    fontname: str
    fontsize: float
    color1: str
    alpha1: str
    color2: str
    alpha2: str
    color3: str
    alpha3: str
    color4: str
    alpha4: str
    bold: bool
    italic: bool
    underline: bool
    strikeout: bool
    scale_x: float
    scale_y: float
    spacing: float
    angle: float
    border_style: bool
    outline: float
    shadow: float
    alignment: int
    margin_l: int
    margin_r: int
    margin_v: int
    encoding: int

    def __repr__(self):
        return pretty_print(self)


class Char:
    """Char object contains informations about a single char of a line in the Ass.

    A char is defined by some text between two karaoke tags (k, ko, kf).

    Attributes:
        i (int): Char index number
        word_i (int): Char word index (e.g.: In line text ``Hello PyonFX users!``, letter "u" will have word_i=2).
        syl_i (int): Char syl index (e.g.: In line text ``{\\k0}Hel{\\k0}lo {\\k0}Pyon{\\k0}FX {\\k0}users!``, letter "F" will have syl_i=3).
        syl_char_i (int): Char invidual syl index (e.g.: In line text ``{\\k0}Hel{\\k0}lo {\\k0}Pyon{\\k0}FX {\\k0}users!``, letter "e" of "users" will have syl_char_i=2).
        start_time (int): Char start time (in milliseconds).
        end_time (int): Char end time (in milliseconds).
        duration (int): Char duration (in milliseconds).
        styleref (obj): Reference to the Style object of this object original line.
        text (str): Char text.
        inline_fx (str): Char inline effect (marked as \\-EFFECT in karaoke-time).
        prespace (int): Char free space before text.
        postspace (int): Char free space after text.
        width (float): Char text width.
        height (float): Char text height.
        x (float): Char text position horizontal (depends on alignment).
        y (float): Char text position vertical (depends on alignment).
        left (float): Char text position left.
        center (float): Char text position center.
        right (float): Char text position right.
        top (float): Char text position top.
        middle (float): Char text position middle.
        bottom (float): Char text position bottom.
    """

    i: int
    word_i: int
    syl_i: int
    syl_char_i: int
    start_time: int
    end_time: int
    duration: int
    styleref: Style
    text: str
    inline_fx: str
    prespace: int
    postspace: int
    width: float
    height: float
    x: float
    y: float
    left: float
    center: float
    right: float
    top: float
    middle: float
    bottom: float

    def __repr__(self):
        return pretty_print(self)


class Syllable:
    """Syllable object contains informations about a single syl of a line in the Ass.

    A syl can be defined as some text after a karaoke tag (k, ko, kf)
    (e.g.: In ``{\\k0}Hel{\\k0}lo {\\k0}Pyon{\\k0}FX {\\k0}users!``, "Pyon" and "FX" are distinct syllables),

    Attributes:
        i (int): Syllable index number
        word_i (int): Syllable word index (e.g.: In line text ``{\\k0}Hel{\\k0}lo {\\k0}Pyon{\\k0}FX {\\k0}users!``, syl "Pyon" will have word_i=1).
        start_time (int): Syllable start time (in milliseconds).
        end_time (int): Syllable end time (in milliseconds).
        duration (int): Syllable duration (in milliseconds).
        styleref (obj): Reference to the Style object of this object original line.
        text (str): Syllable text.
        tags (str): All the remaining tags before syl text apart \\k ones.
        inline_fx (str): Syllable inline effect (marked as \\-EFFECT in karaoke-time).
        prespace (int): Syllable free space before text.
        postspace (int): Syllable free space after text.
        width (float): Syllable text width.
        height (float): Syllable text height.
        x (float): Syllable text position horizontal (depends on alignment).
        y (float): Syllable text position vertical (depends on alignment).
        left (float): Syllable text position left.
        center (float): Syllable text position center.
        right (float): Syllable text position right.
        top (float): Syllable text position top.
        middle (float): Syllable text position middle.
        bottom (float): Syllable text position bottom.
    """

    i: int
    word_i: int
    start_time: int
    end_time: int
    duration: int
    styleref: Style
    text: str
    tags: str
    inline_fx: str
    prespace: int
    postspace: int
    width: float
    height: float
    x: float
    y: float
    left: float
    center: float
    right: float
    top: float
    middle: float
    bottom: float

    def __repr__(self):
        return pretty_print(self)


class Word:
    """Word object contains informations about a single word of a line in the Ass.

    A word can be defined as some text with some optional space before or after
    (e.g.: In the string "What a beautiful world!", "beautiful" and "world" are both distinct words).

    Attributes:
        i (int): Word index number
        start_time (int): Word start time (same as line start time) (in milliseconds).
        end_time (int): Word end time (same as line end time) (in milliseconds).
        duration (int): Word duration (same as line duration) (in milliseconds).
        styleref (obj): Reference to the Style object of this object original line.
        text (str): Word text.
        prespace (int): Word free space before text.
        postspace (int): Word free space after text.
        width (float): Word text width.
        height (float): Word text height.
        x (float): Word text position horizontal (depends on alignment).
        y (float): Word text position vertical (depends on alignment).
        left (float): Word text position left.
        center (float): Word text position center.
        right (float): Word text position right.
        top (float): Word text position top.
        middle (float): Word text position middle.
        bottom (float): Word text position bottom.
    """

    i: int
    start_time: int
    end_time: int
    duration: int
    styleref: Style
    text: str
    prespace: int
    postspace: int
    width: float
    height: float
    x: float
    y: float
    left: float
    center: float
    right: float
    top: float
    middle: float
    bottom: float

    def __repr__(self):
        return pretty_print(self)


class Line:
    """Line object contains informations about a single line in the Ass.

    Note:
        (*) = This field is available only if :class:`extended<Ass>` = True

    Attributes:
        i (int): Line index number
        comment (bool): If *True*, this line will not be displayed on the screen.
        layer (int): Layer for the line. Higher layer numbers are drawn on top of lower ones.
        start_time (int): Line start time (in milliseconds).
        end_time (int): Line end time (in milliseconds).
        duration (int): Line duration (in milliseconds) (*).
        leadin (float): Time between this line and the previous one (in milliseconds; first line = 1000.1) (*).
        leadout (float): Time between this line and the next one (in milliseconds; first line = 1000.1) (*).
        style (str): Style name used for this line.
        styleref (obj): Reference to the Style object of this line (*).
        actor (str): Actor field.
        margin_l (int): Left margin for this line.
        margin_r (int): Right margin for this line.
        margin_v (int): Vertical margin for this line.
        effect (str): Effect field.
        raw_text (str): Line raw text.
        text (str): Line stripped text (no tags).
        width (float): Line text width (*).
        height (float): Line text height (*).
        ascent (float): Line font ascent (*).
        descent (float): Line font descent (*).
        internal_leading (float): Line font internal lead (*).
        external_leading (float): Line font external lead (*).
        x (float): Line text position horizontal (depends on alignment) (*).
        y (float): Line text position vertical (depends on alignment) (*).
        left (float): Line text position left (*).
        center (float): Line text position center (*).
        right (float): Line text position right (*).
        top (float): Line text position top (*).
        middle (float): Line text position middle (*).
        bottom (float): Line text position bottom (*).
        words (list): List containing objects :class:`Word` in this line (*).
        syls (list): List containing objects :class:`Syllable` in this line (if available) (*).
        chars (list): List containing objects :class:`Char` in this line (*).
    """

    i: int
    comment: bool
    layer: int
    start_time: int
    end_time: int
    duration: int
    leadin: float
    leadout: float
    style: str
    styleref: Style
    actor: str
    margin_l: int
    margin_r: int
    margin_v: int
    effect: str
    raw_text: str
    text: str
    width: float
    height: float
    ascent: float
    descent: float
    internal_leading: float
    external_leading: float
    x: float
    y: float
    left: float
    center: float
    right: float
    top: float
    middle: float
    bottom: float
    words: List[Word]
    syls: List[Syllable]
    chars: List[Char]

    def __repr__(self):
        return pretty_print(self)

    def copy(self) -> Line:
        """
        Returns:
            A deep copy of this object (line)
        """
        return copy.deepcopy(self)


class Ass:
    """Contains all the informations about a file in the ASS format and the methods to work with it for both input and output.

    | Usually you will create an Ass object and use it for input and output (see example_ section).
    | PyonFX set automatically an absolute path for all the info in the output, so that wherever you will
      put your generated file, it should always load correctly video and audio.

    Args:
        path_input (str): Path for the input file (either relative to your .py file or absolute).
        path_output (str): Path for the output file (either relative to your .py file or absolute) (DEFAULT: "Output.ass").
        keep_original (bool): If True, you will find all the lines of the input file commented before the new lines generated.
        extended (bool): Calculate more informations from lines (usually you will not have to touch this).
        vertical_kanji (bool): If True, line text with alignment 4, 5 or 6 will be positioned vertically.
            Additionally, ``line`` fields will be re-calculated based on the re-positioned ``line.chars``.

    Attributes:
        path_input (str): Path for input file (absolute).
        path_output (str): Path for output file (absolute).
        meta (:class:`Meta`): Contains informations about the ASS given.
        styles (list of :class:`Style`): Contains all the styles in the ASS given.
        lines (list of :class:`Line`): Contains all the lines (events) in the ASS given.

    .. _example:
    Example:
        ..  code-block:: python3

            io = Ass("in.ass")
            meta, styles, lines = io.get_data()
    """

    def __init__(
        self,
        path_input: str = "",
        path_output: str = "Output.ass",
        keep_original: bool = True,
        extended: bool = True,
        vertical_kanji: bool = False,
    ):
        # Starting to take process time
        self.__saved = False
        self.__plines = 0
        self.__ptime = time.time()

        self.meta, self.styles, self.lines = Meta(), {}, []
        # Getting absolute sub file path
        dirname = os.path.dirname(os.path.abspath(sys.argv[0]))
        if not os.path.isabs(path_input):
            path_input = os.path.join(dirname, path_input)

        # Checking sub file validity (does it exists?)
        if not os.path.isfile(path_input):
            raise FileNotFoundError(
                "Invalid path for the Subtitle file: %s" % path_input
            )

        # Getting absolute output file path
        if path_output == "Output.ass":
            path_output = os.path.join(dirname, path_output)
        elif not os.path.isabs(path_output):
            path_output = os.path.join(dirname, path_output)

        self.path_input = path_input
        self.path_output = path_output
        self.__output = []
        self.__output_extradata = []

        section = ""
        li = 0
        for line in open(self.path_input, "r", encoding="utf-8-sig"):
            # Getting section
            section_pattern = re.compile(r"^\[([^\]]*)")
            if section_pattern.match(line):
                # Updating section
                section = section_pattern.match(line)[1]
                # Appending line to output
                if section != "Aegisub Extradata":
                    self.__output.append(line)

            # Parsing Meta data
            elif section == "Script Info" or section == "Aegisub Project Garbage":
                # Internal function that tries to get the absolute path for media files in meta
                def get_media_abs_path(mediafile):
                    # If this is not a dummy video, let's try to get the absolute path for the video
                    if not mediafile.startswith("?dummy"):
                        tmp = mediafile
                        media_dir = os.path.dirname(self.path_input)

                        while mediafile.startswith("../"):
                            media_dir = os.path.dirname(media_dir)
                            mediafile = mediafile[3:]

                        mediafile = os.path.normpath(
                            "%s%s%s" % (media_dir, os.sep, mediafile)
                        )

                        if not os.path.isfile(mediafile):
                            mediafile = tmp

                    return mediafile

                # Switch
                if re.match(r"WrapStyle: *?(\d+)$", line):
                    self.meta.wrap_style = int(line[11:].strip())
                elif re.match(r"ScaledBorderAndShadow: *?(.+)$", line):
                    self.meta.scaled_border_and_shadow = line[23:].strip() == "yes"
                elif re.match(r"PlayResX: *?(\d+)$", line):
                    self.meta.play_res_x = int(line[10:].strip())
                elif re.match(r"PlayResY: *?(\d+)$", line):
                    self.meta.play_res_y = int(line[10:].strip())
                elif re.match(r"Audio File: *?(.*)$", line):
                    self.meta.audio = get_media_abs_path(line[11:].strip())
                    line = "Audio File: %s\n" % self.meta.audio
                elif re.match(r"Video File: *?(.*)$", line):
                    self.meta.video = get_media_abs_path(line[11:].strip())
                    line = "Video File: %s\n" % self.meta.video

                # Appending line to output
                self.__output.append(line)

            # Parsing Styles
            elif section == "V4+ Styles":
                # Appending line to output
                self.__output.append(line)
                style = re.match(r"Style: (.+?)$", line)

                if style:
                    # Name, Fontname, Fontsize, PrimaryColour, SecondaryColour, OutlineColour, BackColour,
                    # Bold, Italic, Underline, StrikeOut, ScaleX, ScaleY, Spacing, Angle,
                    # BorderStyle, Outline, Shadow, Alignment, MarginL, MarginR, MarginV, Encoding
                    style = [el for el in style[1].split(",")]
                    tmp = Style()

                    tmp.fontname = style[1]
                    tmp.fontsize = float(style[2])

                    tmp.color1, tmp.alpha1 = f"&H{style[3][4:]}&", f"{style[3][:4]}&"
                    tmp.color2, tmp.alpha2 = f"&H{style[4][4:]}&", f"{style[4][:4]}&"
                    tmp.color3, tmp.alpha3 = f"&H{style[5][4:]}&", f"{style[5][:4]}&"
                    tmp.color4, tmp.alpha4 = f"&H{style[6][4:]}&", f"{style[6][:4]}&"

                    tmp.bold = style[7] == "-1"
                    tmp.italic = style[8] == "-1"
                    tmp.underline = style[9] == "-1"
                    tmp.strikeout = style[10] == "-1"

                    tmp.scale_x = float(style[11])
                    tmp.scale_y = float(style[12])

                    tmp.spacing = float(style[13])
                    tmp.angle = float(style[14])

                    tmp.border_style = style[15] == "3"
                    tmp.outline = float(style[16])
                    tmp.shadow = float(style[17])

                    tmp.alignment = int(style[18])
                    tmp.margin_l = int(style[19])
                    tmp.margin_r = int(style[20])
                    tmp.margin_v = int(style[21])

                    tmp.encoding = int(style[22])

                    self.styles[style[0]] = tmp

            # Parsing Dialogues
            elif section == "Events":
                # Appending line to output (commented) if keep_original is True
                if keep_original:
                    self.__output.append(
                        re.sub(r"^(Dialogue|Comment):", "Comment:", line, count=1)
                    )
                elif line.startswith("Format"):
                    self.__output.append(line.strip())

                # Analyzing line
                line = re.match(r"(Dialogue|Comment): (.+?)$", line)

                if line:
                    # Layer, Start, End, Style, Name, MarginL, MarginR, MarginV, Effect, Text
                    tmp = Line()

                    tmp.i = li
                    li += 1

                    tmp.comment = line[1] == "Comment"
                    line = [el for el in line[2].split(",")]

                    tmp.layer = int(line[0])

                    tmp.start_time = Convert.time(line[1])
                    tmp.end_time = Convert.time(line[2])

                    tmp.style = line[3]
                    tmp.actor = line[4]

                    tmp.margin_l = int(line[5])
                    tmp.margin_r = int(line[6])
                    tmp.margin_v = int(line[7])

                    tmp.effect = line[8]

                    tmp.raw_text = ",".join(line[9:])

                    self.lines.append(tmp)

            elif section == "Aegisub Extradata":
                self.__output_extradata.append(line)

            else:
                raise ValueError(f"Unexpected section in the input file: [{section}]")

        # Adding informations to lines and meta?
        if not extended:
            return None

        lines_by_styles = {}
        # Let the fun begin (Pyon!)
        for li, line in enumerate(self.lines):
            try:
                line.styleref = self.styles[line.style]
            except KeyError:
                line.styleref = None

            # Append dialog to styles (for leadin and leadout later)
            if line.style not in lines_by_styles:
                lines_by_styles[line.style] = []
            lines_by_styles[line.style].append(line)

            line.duration = line.end_time - line.start_time
            line.text = re.sub(r"\{.*?\}", "", line.raw_text)

            # Add dialog text sizes and positions (if possible)
            if line.styleref:
                # Creating a Font object and saving return values of font.get_metrics() for the future
                font = Font(line.styleref)
                font_metrics = font.get_metrics()

                line.width, line.height = font.get_text_extents(line.text)
                (
                    line.ascent,
                    line.descent,
                    line.internal_leading,
                    line.external_leading,
                ) = font_metrics

                if self.meta.play_res_x > 0 and self.meta.play_res_y > 0:
                    # Horizontal position
                    tmp_margin_l = (
                        line.margin_l if line.margin_l != 0 else line.styleref.margin_l
                    )
                    tmp_margin_r = (
                        line.margin_r if line.margin_r != 0 else line.styleref.margin_r
                    )

                    if (line.styleref.alignment - 1) % 3 == 0:
                        line.left = tmp_margin_l
                        line.center = line.left + line.width / 2
                        line.right = line.left + line.width
                        line.x = line.left
                    elif (line.styleref.alignment - 2) % 3 == 0:
                        line.left = (
                            self.meta.play_res_x / 2
                            - line.width / 2
                            + tmp_margin_l / 2
                            - tmp_margin_r / 2
                        )
                        line.center = line.left + line.width / 2
                        line.right = line.left + line.width
                        line.x = line.center
                    else:
                        line.left = self.meta.play_res_x - tmp_margin_r - line.width
                        line.center = line.left + line.width / 2
                        line.right = line.left + line.width
                        line.x = line.right

                    # Vertical position
                    if line.styleref.alignment > 6:
                        line.top = (
                            line.margin_v
                            if line.margin_v != 0
                            else line.styleref.margin_v
                        )
                        line.middle = line.top + line.height / 2
                        line.bottom = line.top + line.height
                        line.y = line.top
                    elif line.styleref.alignment > 3:
                        line.top = self.meta.play_res_y / 2 - line.height / 2
                        line.middle = line.top + line.height / 2
                        line.bottom = line.top + line.height
                        line.y = line.middle
                    else:
                        line.top = (
                            self.meta.play_res_y
                            - (
                                line.margin_v
                                if line.margin_v != 0
                                else line.styleref.margin_v
                            )
                            - line.height
                        )
                        line.middle = line.top + line.height / 2
                        line.bottom = line.top + line.height
                        line.y = line.bottom

                # Calculating space width and saving spacing
                space_width = font.get_text_extents(" ")[0]
                style_spacing = line.styleref.spacing

                # Adding words
                line.words = []

                wi = 0
                for prespace, word_text, postspace in re.findall(
                    r"(\s*)([^\s]+)(\s*)", line.text
                ):
                    word = Word()

                    word.i = wi
                    wi += 1

                    word.start_time = line.start_time
                    word.end_time = line.end_time
                    word.duration = line.duration

                    word.styleref = line.styleref
                    word.text = word_text

                    word.prespace = len(prespace)
                    word.postspace = len(postspace)

                    word.width, word.height = font.get_text_extents(word.text)
                    (
                        word.ascent,
                        word.descent,
                        word.internal_leading,
                        word.external_leading,
                    ) = font_metrics

                    line.words.append(word)

                # Calculate word positions with all words data already available
                if line.words and self.meta.play_res_x > 0 and self.meta.play_res_y > 0:
                    if line.styleref.alignment > 6 or line.styleref.alignment < 4:
                        cur_x = line.left
                        for word in line.words:
                            # Horizontal position
                            cur_x = cur_x + word.prespace * (
                                space_width + style_spacing
                            )

                            word.left = cur_x
                            word.center = word.left + word.width / 2
                            word.right = word.left + word.width

                            if (line.styleref.alignment - 1) % 3 == 0:
                                word.x = word.left
                            elif (line.styleref.alignment - 2) % 3 == 0:
                                word.x = word.center
                            else:
                                word.x = word.right

                            # Vertical position
                            word.top = line.top
                            word.middle = line.middle
                            word.bottom = line.bottom
                            word.y = line.y

                            # Updating cur_x
                            cur_x = (
                                cur_x
                                + word.width
                                + word.postspace * (space_width + style_spacing)
                                + style_spacing
                            )
                    else:
                        max_width, sum_height = 0, 0
                        for word in line.words:
                            max_width = max(max_width, word.width)
                            sum_height = sum_height + word.height

                        cur_y = x_fix = self.meta.play_res_y / 2 - sum_height / 2
                        for word in line.words:
                            # Horizontal position
                            x_fix = (max_width - word.width) / 2

                            if line.styleref.alignment == 4:
                                word.left = line.left + x_fix
                                word.center = word.left + word.width / 2
                                word.right = word.left + word.width
                                word.x = word.left
                            elif line.styleref.alignment == 5:
                                word.left = self.meta.play_res_x / 2 - word.width / 2
                                word.center = word.left + word.width / 2
                                word.right = word.left + word.width
                                word.x = word.center
                            else:
                                word.left = line.right - word.width - x_fix
                                word.center = word.left + word.width / 2
                                word.right = word.left + word.width
                                word.x = word.right

                            # Vertical position
                            word.top = cur_y
                            word.middle = word.top + word.height / 2
                            word.bottom = word.top + word.height
                            word.y = word.middle
                            cur_y = cur_y + word.height

                # Search for dialog's text chunks, to later create syllables
                # A text chunk is a text with one or more {tags} preceding it
                # Tags can be some text or empty string
                text_chunks = []
                tag_pattern = re.compile(r"(\{.*?\})+")
                tag = tag_pattern.search(line.raw_text)
                word_i = 0

                if not tag:
                    # No tags found
                    text_chunks.append({"tags": "", "text": line.raw_text})
                else:
                    # First chunk without tags?
                    if tag.start() != 0:
                        text_chunks.append(
                            {"tags": "", "text": line.raw_text[0 : tag.start()]}
                        )

                    # Searching for other tags
                    while True:
                        next_tag = tag_pattern.search(line.raw_text, tag.end())
                        tmp = {
                            # Note that we're removing possibles '}{' caused by consecutive tags
                            "tags": line.raw_text[
                                tag.start() + 1 : tag.end() - 1
                            ].replace("}{", ""),
                            "text": line.raw_text[
                                tag.end() : (next_tag.start() if next_tag else None)
                            ],
                            "word_i": word_i,
                        }
                        text_chunks.append(tmp)

                        # If there are some spaces after text, then we're at the end of the current word
                        if re.match(r"(.*?)(\s+)$", tmp["text"]):
                            word_i = word_i + 1

                        if not next_tag:
                            break
                        tag = next_tag

                # Adding syls
                si = 0
                last_time = 0
                inline_fx = ""
                syl_tags_pattern = re.compile(r"(.*?)\\[kK][of]?(\d+)(.*)")

                line.syls = []
                for tc in text_chunks:
                    # If we don't have at least one \k tag, everything is invalid
                    if not syl_tags_pattern.match(tc["tags"]):
                        line.syls.clear()
                        break

                    posttags = tc["tags"]
                    syls_in_text_chunk = []
                    while True:
                        # Are there \k in posttags?
                        tags_syl = syl_tags_pattern.match(posttags)

                        if not tags_syl:
                            # Append all the temporary syls, except last one
                            for syl in syls_in_text_chunk[:-1]:
                                curr_inline_fx = re.search(r"\\\-([^\\]+)", syl.tags)
                                if curr_inline_fx:
                                    inline_fx = curr_inline_fx[1]
                                syl.inline_fx = inline_fx

                                # Hidden syls are treated like empty syls
                                syl.prespace, syl.text, syl.postspace = 0, "", 0

                                syl.width, syl.height = font.get_text_extents("")
                                (
                                    syl.ascent,
                                    syl.descent,
                                    syl.internal_leading,
                                    syl.external_leading,
                                ) = font_metrics

                                line.syls.append(syl)

                            # Append last syl
                            syl = syls_in_text_chunk[-1]
                            syl.tags += posttags

                            curr_inline_fx = re.search(r"\\\-([^\\]+)", syl.tags)
                            if curr_inline_fx:
                                inline_fx = curr_inline_fx[1]
                            syl.inline_fx = inline_fx

                            if tc["text"].isspace():
                                syl.prespace, syl.text, syl.postspace = 0, tc["text"], 0
                            else:
                                syl.prespace, syl.text, syl.postspace = re.match(
                                    r"(\s*)(.*?)(\s*)$", tc["text"]
                                ).groups()
                                syl.prespace, syl.postspace = (
                                    len(syl.prespace),
                                    len(syl.postspace),
                                )

                            syl.width, syl.height = font.get_text_extents(syl.text)
                            (
                                syl.ascent,
                                syl.descent,
                                syl.internal_leading,
                                syl.external_leading,
                            ) = font_metrics

                            line.syls.append(syl)
                            break

                        pretags, kdur, posttags = tags_syl.groups()

                        # Create a Syllable object
                        syl = Syllable()

                        syl.start_time = last_time
                        syl.end_time = last_time + int(kdur) * 10
                        syl.duration = int(kdur) * 10

                        syl.styleref = line.styleref
                        syl.tags = pretags

                        syl.i = si
                        syl.word_i = tc["word_i"]

                        syls_in_text_chunk.append(syl)

                        # Update working variable
                        si += 1
                        last_time = syl.end_time

                # Calculate syllables positions with all syllables data already available
                if line.syls and self.meta.play_res_x > 0 and self.meta.play_res_y > 0:
                    if (
                        line.styleref.alignment > 6
                        or line.styleref.alignment < 4
                        or not vertical_kanji
                    ):
                        cur_x = line.left
                        for syl in line.syls:
                            cur_x = cur_x + syl.prespace * (space_width + style_spacing)
                            # Horizontal position
                            syl.left = cur_x
                            syl.center = syl.left + syl.width / 2
                            syl.right = syl.left + syl.width

                            if (line.styleref.alignment - 1) % 3 == 0:
                                syl.x = syl.left
                            elif (line.styleref.alignment - 2) % 3 == 0:
                                syl.x = syl.center
                            else:
                                syl.x = syl.right

                            cur_x = (
                                cur_x
                                + syl.width
                                + syl.postspace * (space_width + style_spacing)
                                + style_spacing
                            )

                            # Vertical position
                            syl.top = line.top
                            syl.middle = line.middle
                            syl.bottom = line.bottom
                            syl.y = line.y

                    else:  # Kanji vertical position
                        max_width, sum_height = 0, 0
                        for syl in line.syls:
                            max_width = max(max_width, syl.width)
                            sum_height = sum_height + syl.height

                        cur_y = self.meta.play_res_y / 2 - sum_height / 2

                        for syl in line.syls:
                            # Horizontal position
                            x_fix = (max_width - syl.width) / 2
                            if line.styleref.alignment == 4:
                                syl.left = line.left + x_fix
                                syl.center = syl.left + syl.width / 2
                                syl.right = syl.left + syl.width
                                syl.x = syl.left
                            elif line.styleref.alignment == 5:
                                syl.left = line.center - syl.width / 2
                                syl.center = syl.left + syl.width / 2
                                syl.right = syl.left + syl.width
                                syl.x = syl.center
                            else:
                                syl.left = line.right - syl.width - x_fix
                                syl.center = syl.left + syl.width / 2
                                syl.right = syl.left + syl.width
                                syl.x = syl.right

                            # Vertical position
                            syl.top = cur_y
                            syl.middle = syl.top + syl.height / 2
                            syl.bottom = syl.top + syl.height
                            syl.y = syl.middle
                            cur_y = cur_y + syl.height

                # Adding chars
                line.chars = []

                # If we have syls in line, we prefert to work with them to provide more informations
                if line.syls:
                    words_or_syls = line.syls
                else:
                    words_or_syls = line.words

                # Getting chars
                char_index = 0
                for el in words_or_syls:
                    el_text = "{}{}{}".format(
                        " " * el.prespace, el.text, " " * el.postspace
                    )
                    for ci, char_text in enumerate(list(el_text)):
                        char = Char()
                        char.i = ci

                        # If we're working with syls, we can add some indexes
                        char.i = char_index
                        char_index += 1
                        if line.syls:
                            char.word_i = el.word_i
                            char.syl_i = el.i
                            char.syl_char_i = ci
                        else:
                            char.word_i = el.i

                        # Adding last fields based on the existance of syls or not
                        char.start_time = el.start_time
                        char.end_time = el.end_time
                        char.duration = el.duration

                        char.styleref = line.styleref
                        char.text = char_text

                        char.width, char.height = font.get_text_extents(char.text)
                        (
                            char.ascent,
                            char.descent,
                            char.internal_leading,
                            char.external_leading,
                        ) = font_metrics

                        line.chars.append(char)

                # Calculate character positions with all characters data already available
                if line.chars and self.meta.play_res_x > 0 and self.meta.play_res_y > 0:
                    if (
                        line.styleref.alignment > 6
                        or line.styleref.alignment < 4
                        or not vertical_kanji
                    ):
                        cur_x = line.left
                        for char in line.chars:
                            # Horizontal position
                            char.left = cur_x
                            char.center = char.left + char.width / 2
                            char.right = char.left + char.width

                            if (line.styleref.alignment - 1) % 3 == 0:
                                char.x = char.left
                            elif (line.styleref.alignment - 2) % 3 == 0:
                                char.x = char.center
                            else:
                                char.x = char.right

                            cur_x = cur_x + char.width + style_spacing

                            # Vertical position
                            char.top = line.top
                            char.middle = line.middle
                            char.bottom = line.bottom
                            char.y = line.y
                    else:
                        max_width, sum_height = 0, 0
                        for char in line.chars:
                            max_width = max(max_width, char.width)
                            sum_height = sum_height + char.height

                        cur_y = x_fix = self.meta.play_res_y / 2 - sum_height / 2

                        # Fixing line positions
                        line.top = cur_y
                        line.middle = self.meta.play_res_y / 2
                        line.bottom = line.top + sum_height
                        line.width = max_width
                        line.height = sum_height
                        if line.styleref.alignment == 4:
                            line.center = line.left + max_width / 2
                            line.right = line.left + max_width
                        elif line.styleref.alignment == 5:
                            line.left = line.center - max_width / 2
                            line.right = line.left + max_width
                        else:
                            line.left = line.right - max_width
                            line.center = line.left + max_width / 2

                        for char in line.chars:
                            # Horizontal position
                            x_fix = (max_width - char.width) / 2
                            if line.styleref.alignment == 4:
                                char.left = line.left + x_fix
                                char.center = char.left + char.width / 2
                                char.right = char.left + char.width
                                char.x = char.left
                            elif line.styleref.alignment == 5:
                                char.left = self.meta.play_res_x / 2 - char.width / 2
                                char.center = char.left + char.width / 2
                                char.right = char.left + char.width
                                char.x = char.center
                            else:
                                char.left = line.right - char.width - x_fix
                                char.center = char.left + char.width / 2
                                char.right = char.left + char.width
                                char.x = char.right

                            # Vertical position
                            char.top = cur_y
                            char.middle = char.top + char.height / 2
                            char.bottom = char.top + char.height
                            char.y = char.middle
                            cur_y = cur_y + char.height

        # Add durations between dialogs
        for style in lines_by_styles:
            lines_by_styles[style].sort(key=lambda x: x.start_time)
            for li, line in enumerate(lines_by_styles[style]):
                line.leadin = (
                    1000.1
                    if li == 0
                    else line.start_time - lines_by_styles[style][li - 1].end_time
                )
                line.leadout = (
                    1000.1
                    if li == len(lines_by_styles[style]) - 1
                    else lines_by_styles[style][li + 1].start_time - line.end_time
                )

    def get_data(self) -> Tuple[Meta, Style, List[Line]]:
        """Utility function to retrieve easily meta styles and lines.

        Returns:
            :attr:`meta`, :attr:`styles` and :attr:`lines`
        """
        return self.meta, self.styles, self.lines

    def write_line(self, line: Line) -> Optional[TypeError]:
        """Appends a line to the output list (which is private) that later on will be written to the output file when calling save().

        Use it whenever you've prepared a line, it will not impact performance since you
        will not actually write anything until :func:`save` will be called.

        Parameters:
            line (:class:`Line`): A line object. If not valid, TypeError is raised.
        """
        if isinstance(line, Line):
            self.__output.append(
                "\n%s: %d,%s,%s,%s,%s,%04d,%04d,%04d,%s,%s"
                % (
                    "Comment" if line.comment else "Dialogue",
                    line.layer,
                    Convert.time(max(0, int(line.start_time))),
                    Convert.time(max(0, int(line.end_time))),
                    line.style,
                    line.actor,
                    line.margin_l,
                    line.margin_r,
                    line.margin_v,
                    line.effect,
                    line.text,
                )
            )
            self.__plines += 1
        else:
            raise TypeError("Expected Line object, got %s." % type(line))

    def save(self, quiet: bool = False) -> None:
        """Write everything inside the private output list to a file.

        Parameters:
            quiet (bool): If True, you will not get printed any message.
        """

        # Writing to file
        with open(self.path_output, "w", encoding="utf-8-sig") as f:
            f.writelines(self.__output + ["\n"])
            if self.__output_extradata:
                f.write("\n[Aegisub Extradata]\n")
                f.writelines(self.__output_extradata)

        self.__saved = True

        if not quiet:
            print(
                "Produced lines: %d\nProcess duration (in seconds): %.3f"
                % (self.__plines, time.time() - self.__ptime)
            )

    def open_aegisub(self) -> int:
        """Open the output (specified in self.path_output) with Aegisub.

        This can be usefull if you don't have MPV installed or you want to look at your output in detailed.

        Returns:
            0 if success, -1 if the output couldn't be opened.
        """

        # Check if it was saved
        if not self.__saved:
            print(
                "[WARNING] You've tried to open the output with Aegisub before having saved. Check your code."
            )
            return -1

        if sys.platform == "win32":
            os.startfile(self.path_output)
        else:
            try:
                subprocess.call(["aegisub", os.path.abspath(self.path_output)])
            except FileNotFoundError:
                print("[WARNING] Aegisub not found.")
                return -1

        return 0

    def open_mpv(
        self, video_path: str = "", video_start: str = "", full_screen: bool = False
    ) -> int:
        """Open the output (specified in self.path_output) in softsub with the MPV player.
        To utilize this function, MPV player is required. Additionally if you're on Windows, MPV must be in the PATH (check https://pyonfx.readthedocs.io/en/latest/quick%20start.html#installation-extra-step).

        This is one of the fastest way to reproduce your output in a comfortable way.

        Parameters:
            video_path (string): The video file path (absolute) to reproduce. If not specified, **meta.video** is automatically taken.
            video_start (string): The start time for the video (more info: https://mpv.io/manual/master/#options-start). If not specified, 0 is automatically taken.
            full_screen (bool): If True, it will reproduce the output in full screen. If not specified, False is automatically taken.
        """

        # Check if it was saved
        if not self.__saved:
            print(
                "[ERROR] You've tried to open the output with MPV before having saved. Check your code."
            )
            return -1

        # Check if mpv is usable
        if self.meta.video.startswith("?dummy") and not video_path:
            print(
                "[WARNING] Cannot use MPV (if you have it in your PATH) for file preview, since your .ass contains a dummy video.\n"
                "You can specify a new video source using video_path parameter, check the documentation of the function."
            )
            return -1

        # Setting up the command to execute
        cmd = ["mpv"]

        if not video_path:
            cmd.append(self.meta.video)
        else:
            cmd.append(video_path)
        if video_start:
            cmd.append("--start=" + video_start)
        if full_screen:
            cmd.append("--fs")

        cmd.append("--sub-file=" + self.path_output)

        try:
            subprocess.call(cmd)
        except FileNotFoundError:
            print(
                "[WARNING] MPV not found in your environment variables.\n"
                "Please refer to the documentation's \"Quick Start\" section if you don't know how to solve it."
            )
            return -1

        return 0


def pretty_print(
    obj: Union[Meta, Style, Line, Word, Syllable, Char], indent: int = 0, name: str = ""
) -> str:
    # Utility function to print object Meta, Style, Line, Word, Syllable and Char (this is a dirty solution probably)
    if type(obj) == Line:
        out = " " * indent + f"lines[{obj.i}] ({type(obj).__name__}):\n"
    elif type(obj) == Word:
        out = " " * indent + f"words[{obj.i}] ({type(obj).__name__}):\n"
    elif type(obj) == Syllable:
        out = " " * indent + f"syls[{obj.i}] ({type(obj).__name__}):\n"
    elif type(obj) == Char:
        out = " " * indent + f"chars[{obj.i}] ({type(obj).__name__}):\n"
    else:
        out = " " * indent + f"{name}({type(obj).__name__}):\n"

    # Let's print all this object fields
    indent += 4
    for k, v in obj.__dict__.items():
        if "__dict__" in dir(v):
            # Work recursively to print another object
            out += pretty_print(v, indent, k + " ")
        elif type(v) == list:
            for i, el in enumerate(v):
                # Work recursively to print other objects inside a list
                out += pretty_print(el, indent, f"{k}[{i}] ")
        else:
            # Just print a field of this object
            out += " " * indent + f"{k}: {str(v)}\n"

    return out



# -*- coding: utf-8 -*-
# PyonFX: An easy way to create KFX (Karaoke Effects) and complex typesetting using the ASS format (Advanced Substation Alpha).
# Copyright (C) 2019 Antonio Strippoli (CoffeeStraw/YellowFlash)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyonFX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

from __future__ import annotations
import math
from typing import Callable, Optional, List, Tuple, Union
from pyquaternion import Quaternion
from inspect import signature


class Shape:
    """
    This class can be used to define a Shape object (by passing its drawing commands)
    and then apply functions to it in order to accomplish some tasks, like analyzing its bounding box, apply transformations, splitting curves into segments...

    Args:
        drawing_cmds (str): The shape's drawing commands in ASS format as a string.
    """

    def __init__(self, drawing_cmds: str):
        # Assure that drawing_cmds is a string
        if not isinstance(drawing_cmds, str):
            raise TypeError(
                f"A string containing the shape's drawing commands is expected, but you passed a {type(drawing_cmds)}"
            )
        self.drawing_cmds = drawing_cmds

    def __repr__(self):
        # We return drawing commands as a string rapresentation of the object
        return self.drawing_cmds

    def __eq__(self, other: Shape):
        # Method used to compare two shapes
        return type(other) is type(self) and self.drawing_cmds == other.drawing_cmds

    @staticmethod
    def format_value(x: float, prec: int = 3) -> str:
        # Utility function to properly format values for shapes also returning them as a string
        return f"{x:.{prec}f}".rstrip("0").rstrip(".")

    def has_error(self) -> Union[bool, str]:
        """Utility function that checks if the shape is valid.

        Returns:
            False if no error has been found, else a string with the first error encountered.
        """
        # Obtain commands and points
        cad = self.drawing_cmds.split()
        n = len(cad)
        mode = ""

        # Prepare usefull lists
        two_args_cmds = ["m", "n", "l", "p"]
        six_args_cmds = ["b", "s"]

        # Iterate over commands and points
        i = 0
        while i < n:
            if cad[i] in two_args_cmds:
                # Check if we have an unexpected and of the shape
                if n - i < 3:
                    return (
                        f"Unexpected end of shape ('{cad[i]}' expect at least two args)"
                    )

                # Check if we have two numeric values after command
                try:
                    float(cad[i + 1])
                    float(cad[i + 2])
                except ValueError:
                    return (
                        f"Expected numeric value at: '{cad[i]} {cad[i+1]} {cad[i+2]}'"
                    )

                # Valid, go on
                mode = cad[i]
                i += 3
            elif cad[i] in six_args_cmds:
                # Check if we have an unexpected and of the shape
                if n - i < 7:
                    return (
                        f"Unexpected end of shape ('{cad[i]}' expect at least six args)"
                    )

                # Check if we have six numeric values after command
                try:
                    for x in range(1, 7):
                        float(cad[i + x])
                except ValueError:
                    return f"Expected numeric value at: '{cad[i]} {' '.join(cad[i+1:i+7])}'"

                # Valid, go on
                mode = cad[i]
                i += 7
            elif cad[i] == "c":
                # 'c' expect no arguments, skip
                mode = ""
                i += 1
            elif mode in two_args_cmds:
                # Check if we have an unexpected and of the shape
                if n - i < 2:
                    return (
                        f"Unexpected end of shape ('{cad[i]}' expect at least two args)"
                    )

                # Check if we have two numeric values
                try:
                    float(cad[i])
                    float(cad[i + 1])
                except ValueError:
                    return (
                        f"Expected numeric value at: '{mode} ... {cad[i]} {cad[i+1]}'"
                    )

                # Valid, go on
                i += 2
            elif mode in six_args_cmds:
                # Check if we have an unexpected and of the shape
                if n - i < 6:
                    return (
                        f"Unexpected end of shape ('{cad[i]}' expect at least six args)"
                    )

                # Check if we have six numeric values
                try:
                    for x in range(6):
                        float(cad[i + x])
                except ValueError:
                    return f"Expected numeric value at: '{mode} ... {' '.join(cad[i:i+6])}'"

                # Valid, go on
                i += 6
            else:
                # Wtf is this?
                return f"Unexpected command '{cad[i]}'"

        return False

    def map(
        self, fun: Callable[[float, float, Optional[str]], Tuple[float, float]]
    ) -> Shape:
        """Sends every point of a shape through given transformation function to change them.

        **Tips:** *Working with outline points can be used to deform the whole shape and make f.e. a wobble effect.*

        Parameters:
            fun (function): A function with two (or optionally three) parameters. It will define how each coordinate will be changed. The first two parameters represent the x and y coordinates of each point. The third optional it represents the type of each point (move, line, bezier...).

        Returns:
            A pointer to the current object.

        Examples:
            ..  code-block:: python3

                original = Shape("m 0 0 l 20 0 20 10 0 10")
                print ( original.map(lambda x, y: (x+10, y+5) ) )

            >>> m 10 5 l 30 5 30 15 10 15
        """
        if not callable(fun):
            raise TypeError("(Lambda) function expected")

        # Getting all points and commands in a list
        cmds_and_points = self.drawing_cmds.split()
        i = 0
        n = len(cmds_and_points)

        # Checking whether the function take the typ parameter or not
        if len(signature(fun).parameters) == 2:
            while i < n:
                try:
                    # Applying transformation
                    x, y = fun(float(cmds_and_points[i]), float(cmds_and_points[i + 1]))
                except TypeError:
                    # Values weren't returned, so we don't need to modify them
                    i += 2
                    continue
                except ValueError:
                    # We have found a string, let's skip this
                    i += 1
                    continue
                except IndexError:
                    raise ValueError("Unexpected end of the shape")

                # Convert back to string the results for later
                cmds_and_points[i : i + 2] = (
                    Shape.format_value(x),
                    Shape.format_value(y),
                )
                i += 2
        else:
            typ = ""
            while i < n:
                try:
                    # Applying transformation
                    x, y = fun(
                        float(cmds_and_points[i]), float(cmds_and_points[i + 1]), typ
                    )
                except TypeError:
                    # Values weren't returned, so we don't need to modify them
                    i += 2
                    continue
                except ValueError:
                    # We have found a string, let's skip this
                    typ = cmds_and_points[i]
                    i += 1
                    continue
                except IndexError:
                    raise ValueError("Unexpected end of the shape")

                # Convert back to string the results for later
                cmds_and_points[i : i + 2] = (
                    Shape.format_value(x),
                    Shape.format_value(y),
                )
                i += 2

        # Sew up everything back and update shape
        self.drawing_cmds = " ".join(cmds_and_points)
        return self

    def bounding(self) -> Tuple[float, float, float, float]:
        """Calculates shape bounding box.

        **Tips:** *Using this you can get more precise information about a shape (width, height, position).*

        Returns:
            A tuple (x0, y0, x1, y1) containing coordinates of the bounding box.

        Examples:
            ..  code-block:: python3

                print("Left-top: %d %d\\nRight-bottom: %d %d" % ( Shape("m 10 5 l 25 5 25 42 10 42").bounding() ) )

            >>> Left-top: 10 5
            >>> Right-bottom: 25 42
        """

        # Bounding data
        x0: float = None
        y0: float = None
        x1: float = None
        y1: float = None

        # Calculate minimal and maximal coordinates
        def compute_edges(x, y):
            nonlocal x0, y0, x1, y1
            if x0 is not None:
                x0, y0, x1, y1 = min(x0, x), min(y0, y), max(x1, x), max(y1, y)
            else:
                x0, y0, x1, y1 = x, y, x, y
            return x, y

        self.map(compute_edges)
        return x0, y0, x1, y1

    def move(self, x: float = None, y: float = None) -> Shape:
        """Moves shape coordinates in given direction.

        | If neither x and y are passed, it will automatically center the shape to the origin (0,0).
        | This function is an high level function, it just uses Shape.map, which is more advanced. Additionally, it is an easy way to center a shape.

        Parameters:
            x (int or float): Displacement along the x-axis.
            y (int or float): Displacement along the y-axis.

        Returns:
            A pointer to the current object.

        Examples:
            ..  code-block:: python3

                print( Shape("m 0 0 l 30 0 30 20 0 20").move(-5, 10) )

            >>> m -5 10 l 25 10 25 30 -5 30
        """
        if x is None and y is None:
            x, y = [-1 * el for el in self.bounding()[0:2]]
        elif x is None:
            x = 0
        elif y is None:
            y = 0

        # Update shape
        self.map(lambda cx, cy: (cx + x, cy + y))
        return self

    def flatten(self, tolerance: float = 1.0) -> Shape:
        """Splits shape's bezier curves into lines.

        | This is a low level function. Instead, you should use :func:`split` which already calls this function.

        Parameters:
            tolerance (float): Angle in degree to define a curve as flat (increasing it will boost performance during reproduction, but lower accuracy)

        Returns:
            A pointer to the current object.

        Returns:
            The shape as a string, with bezier curves converted to lines.
        """
        # TO DO: Make this function iterative, recursion is bad.
        if tolerance < 0:
            raise ValueError("Tolerance must be a positive number")

        # Inner functions definitions
        # 4th degree curve subdivider (De Casteljau)
        def curve4_subdivide(
            x0,
            y0,
            x1,
            y1,
            x2,
            y2,
            x3,
            y3,
            pct,
        ):
            # Calculate points on curve vectors
            x01, y01, x12, y12, x23, y23 = (
                (x0 + x1) * pct,
                (y0 + y1) * pct,
                (x1 + x2) * pct,
                (y1 + y2) * pct,
                (x2 + x3) * pct,
                (y2 + y3) * pct,
            )
            x012, y012, x123, y123 = (
                (x01 + x12) * pct,
                (y01 + y12) * pct,
                (x12 + x23) * pct,
                (y12 + y23) * pct,
            )
            x0123, y0123 = (x012 + x123) * pct, (y012 + y123) * pct
            # Return new 2 curves
            return (
                x0,
                y0,
                x01,
                y01,
                x012,
                y012,
                x0123,
                y0123,
                x0123,
                y0123,
                x123,
                y123,
                x23,
                y23,
                x3,
                y3,
            )

        # Check flatness of 4th degree curve with angles
        def curve4_is_flat(
            x0,
            y0,
            x1,
            y1,
            x2,
            y2,
            x3,
            y3,
        ):
            # Pack curve vectors (only ones non zero)
            vecs = [[x1 - x0, y1 - y0], [x2 - x1, y2 - y1], [x3 - x2, y3 - y2]]
            vecs = [el for el in vecs if not (el[0] == 0 and el[1] == 0)]

            # Inner functions to calculate degrees between two 2d vectors
            def dotproduct(v1, v2):
                return sum((a * b) for a, b in zip(v1, v2))

            def length(v):
                return math.sqrt(dotproduct(v, v))

            def get_angle(v1, v2):
                calc = max(
                    min(dotproduct(v1, v2) / (length(v1) * length(v2)), 1), -1
                )  # Clamping value to prevent errors
                angle = math.degrees(math.acos(calc))
                if (v1[0] * v2[1] - v1[1] * v2[0]) < 0:
                    return -angle
                return angle

            # Check flatness on vectors
            for i in range(1, len(vecs)):
                if abs(get_angle(vecs[i - 1], vecs[i])) > tolerance:
                    return False
            return True

        # Inner function to convert 4th degree curve to line points
        def curve4_to_lines(
            x0,
            y0,
            x1,
            y1,
            x2,
            y2,
            x3,
            y3,
        ):
            # Line points buffer
            pts = ""

            # Conversion in recursive processing
            def convert_recursive(x0, y0, x1, y1, x2, y2, x3, y3):
                if curve4_is_flat(x0, y0, x1, y1, x2, y2, x3, y3):
                    nonlocal pts
                    x3, y3 = Shape.format_value(x3), Shape.format_value(y3)
                    pts += f"{x3} {y3} "
                    return

                (
                    x10,
                    y10,
                    x11,
                    y11,
                    x12,
                    y12,
                    x13,
                    y13,
                    x20,
                    y20,
                    x21,
                    y21,
                    x22,
                    y22,
                    x23,
                    y23,
                ) = curve4_subdivide(x0, y0, x1, y1, x2, y2, x3, y3, 0.5)
                convert_recursive(x10, y10, x11, y11, x12, y12, x13, y13)
                convert_recursive(x20, y20, x21, y21, x22, y22, x23, y23)

            # Splitting curve recursively until we're not satisfied (angle <= tolerance)
            convert_recursive(x0, y0, x1, y1, x2, y2, x3, y3)
            # Return resulting points
            return " ".join(
                pts[:-1].split(" ")[:-2]
            )  # Delete last space and last two float values

        # Getting all points and commands in a list
        cmds_and_points = self.drawing_cmds.split()
        i = 0
        n = len(cmds_and_points)

        # Scanning all commands and points (improvable)
        while i < n:
            if (
                cmds_and_points[i] == "b"
            ):  # We've found a curve, let's split it into lines
                try:
                    # Getting all the points: if we don't have exactly 8 points, shape is not valid
                    x0, y0 = (
                        float(cmds_and_points[i - 2]),
                        float(cmds_and_points[i - 1]),
                    )
                    x1, y1 = (
                        float(cmds_and_points[i + 1]),
                        float(cmds_and_points[i + 2]),
                    )
                    x2, y2 = (
                        float(cmds_and_points[i + 3]),
                        float(cmds_and_points[i + 4]),
                    )
                    x3, y3 = (
                        float(cmds_and_points[i + 5]),
                        float(cmds_and_points[i + 6]),
                    )
                except IndexError:
                    raise ValueError(
                        "Shape providen is not valid (not enough points for a curve)"
                    )

                # Obtaining the converted curve and saving it for later
                cmds_and_points[i] = "l"
                cmds_and_points[i + 1] = curve4_to_lines(x0, y0, x1, y1, x2, y2, x3, y3)

                i += 2
                n -= 3

                # Deleting the remaining points
                for _ in range(3):
                    del cmds_and_points[i]

                # Going to the next point
                i += 2

                # Check if we're at the end of the shape
                if i < n:
                    # Check for implicit bezier curve
                    try:
                        float(cmds_and_points[i])  # Next number is a float?
                        cmds_and_points.insert(i, "b")
                        n += 1
                    except ValueError:
                        pass
            elif cmds_and_points[i] == "c":  # Deleting c tag?
                del cmds_and_points[i]
                n -= 1
            else:
                i += 1

        # Update shape
        self.drawing_cmds = " ".join(cmds_and_points)
        return self

    def split(self, max_len: float = 16, tolerance: float = 1.0) -> Shape:
        """Splits shape bezier curves into lines and splits lines into shorter segments with maximum given length.

        **Tips:** *You can call this before using :func:`map` to work with more outline points for smoother deforming.*

        Parameters:
            max_len (int or float): The max length that you want all the lines to be
            tolerance (float): Angle in degree to define a bezier curve as flat (increasing it will boost performance during reproduction, but lower accuracy)

        Returns:
            A pointer to the current object.

        Examples:
            ..  code-block:: python3

                print( Shape("m -100.5 0 l 100 0 b 100 100 -100 100 -100.5 0 c").split() )

            >>> m -100.5 0 l -100 0 -90 0 -80 0 -70 0 -60 0 -50 0 -40 0 -30 0 -20 0 -10 0 0 0 10 0 20 0 30 0 40 0 50 0 60 0 70 0 80 0 90 0 100 0 l 99.964 2.325 99.855 4.614 99.676 6.866 99.426 9.082 99.108 11.261 98.723 13.403 98.271 15.509 97.754 17.578 97.173 19.611 96.528 21.606 95.822 23.566 95.056 25.488 94.23 27.374 93.345 29.224 92.403 31.036 91.405 32.812 90.352 34.552 89.246 36.255 88.086 37.921 86.876 39.551 85.614 41.144 84.304 42.7 82.945 44.22 81.54 45.703 80.088 47.15 78.592 48.56 77.053 49.933 75.471 51.27 73.848 52.57 72.184 53.833 70.482 55.06 68.742 56.25 66.965 57.404 65.153 58.521 63.307 59.601 61.427 60.645 59.515 61.652 57.572 62.622 55.599 63.556 53.598 64.453 51.569 65.314 49.514 66.138 47.433 66.925 45.329 67.676 43.201 68.39 41.052 69.067 38.882 69.708 36.692 70.312 34.484 70.88 32.259 71.411 27.762 72.363 23.209 73.169 18.61 73.828 13.975 74.341 9.311 74.707 4.629 74.927 -0.062 75 -4.755 74.927 -9.438 74.707 -14.103 74.341 -18.741 73.828 -23.343 73.169 -27.9 72.363 -32.402 71.411 -34.63 70.88 -36.841 70.312 -39.033 69.708 -41.207 69.067 -43.359 68.39 -45.49 67.676 -47.599 66.925 -49.683 66.138 -51.743 65.314 -53.776 64.453 -55.782 63.556 -57.759 62.622 -59.707 61.652 -61.624 60.645 -63.509 59.601 -65.361 58.521 -67.178 57.404 -68.961 56.25 -70.707 55.06 -72.415 53.833 -74.085 52.57 -75.714 51.27 -77.303 49.933 -78.85 48.56 -80.353 47.15 -81.811 45.703 -83.224 44.22 -84.59 42.7 -85.909 41.144 -87.178 39.551 -88.397 37.921 -89.564 36.255 -90.68 34.552 -91.741 32.812 -92.748 31.036 -93.699 29.224 -94.593 27.374 -95.428 25.488 -96.205 23.566 -96.92 21.606 -97.575 19.611 -98.166 17.578 -98.693 15.509 -99.156 13.403 -99.552 11.261 -99.881 9.082 -100.141 6.866 -100.332 4.614 -100.452 2.325 -100.5 0
        """
        if max_len <= 0:
            raise ValueError(
                "The length of segments must be a positive and non-zero value"
            )

        # Internal function to help splitting a line
        def line_split(x0: float, y0: float, x1: float, y1: float):
            x0, y0, x1, y1 = float(x0), float(y0), float(x1), float(y1)
            # Line direction & length
            rel_x, rel_y = x1 - x0, y1 - y0
            distance = math.sqrt(rel_x * rel_x + rel_y * rel_y)
            # If the line is too long -> split
            if distance > max_len:
                lines: list[str] = []
                distance_rest = distance % max_len
                cur_distance = distance_rest if distance_rest > 0 else max_len

                while cur_distance <= distance:
                    pct = cur_distance / distance
                    x, y = (
                        Shape.format_value(x0 + rel_x * pct),
                        Shape.format_value(y0 + rel_y * pct),
                    )

                    lines.append(f"{x} {y}")
                    cur_distance += max_len

                return " ".join(lines), lines[-1].split()
            else:  # No line split
                x1, y1 = Shape.format_value(x1), Shape.format_value(y1)
                return f"{x1} {y1}", [x1, y1]

        # Getting all points and commands in a list
        cmds_and_points = self.flatten().drawing_cmds.split()
        i = 0
        n = len(cmds_and_points)

        # Utility variables
        is_line = False
        previous_two = None
        last_move = None

        # Splitting everything splittable, probably improvable
        while i < n:
            current = cmds_and_points[i]
            if current == "l":
                # Activate line mode, save previous two points
                is_line = True
                if (
                    not previous_two
                ):  # If we're not running into contiguous line, we need to save the previous two
                    previous_two = [cmds_and_points[i - 2], cmds_and_points[i - 1]]
                i += 1
            elif (
                current == "m"
                or current == "n"
                or current == "b"
                or current == "s"
                or current == "p"
                or current == "c"
            ):
                if current == "m":
                    if (
                        last_move
                    ):  # If we had a previous move, we need to close the previous figure before proceding
                        x0, y0 = None, None
                        if (
                            previous_two
                        ):  # If I don't have previous point, I can read them on cmds_and_points, else I wil take 'em
                            x0, y0 = previous_two[0], previous_two[1]
                        else:
                            x0, y0 = cmds_and_points[i - 2], cmds_and_points[i - 1]

                        if not (
                            x0 == last_move[0] and y0 == last_move[1]
                        ):  # Closing last figure
                            cmds_and_points[i] = (
                                line_split(x0, y0, last_move[0], last_move[1])[0] + " m"
                            )
                    last_move = [cmds_and_points[i + 1], cmds_and_points[i + 2]]

                # Disabling line mode, removing previous two points
                is_line = False
                previous_two = None
                i += 1
            elif is_line:
                # Do the work with the two points found and the previous two
                cmds_and_points[i], previous_two = line_split(
                    previous_two[0],
                    previous_two[1],
                    cmds_and_points[i],
                    cmds_and_points[i + 1],
                )
                del cmds_and_points[i + 1]
                # Let's go to the next two points or tag
                i += 1
                n -= 1
            else:  # We're working with points that are not lines points, let's go forward
                i += 2

        # Close last figure of new shape, taking two last points and two last points of move
        i = n
        if not previous_two:
            while i >= 0:
                current = cmds_and_points[i]
                current_prev = cmds_and_points[i - 1]
                if (
                    current != "m"
                    and current != "n"
                    and current != "b"
                    and current != "s"
                    and current != "p"
                    and current != "c"
                    and current_prev != "m"
                    and current_prev != "n"
                    and current_prev != "b"
                    and current_prev != "s"
                    and current_prev != "p"
                    and current_prev != "c"
                ):
                    previous_two = [current, current_prev]
                    break
                i -= 1
        if not (
            previous_two[0] == last_move[0] and previous_two[1] == last_move[1]
        ):  # Split!
            cmds_and_points.append(
                "l "
                + line_split(
                    previous_two[0], previous_two[1], last_move[0], last_move[1]
                )[0]
            )

        # Sew up everything back and update shape
        self.drawing_cmds = " ".join(cmds_and_points)
        return self

    def __to_outline(
        self, bord_xy: float, bord_y: float = None, mode: str = "round"
    ) -> Shape:
        """Converts shape command for filling to a shape command for stroking.

        **Tips:** *You could use this for border textures.*

        Parameters:
            shape (str): The shape in ASS format as a string.

        Returns:
            A pointer to the current object.

        Returns:
            A new shape as string, representing the border of the input.
        """
        raise NotImplementedError

    @staticmethod
    def ring(out_r: float, in_r: float) -> Shape:
        """Returns a shape object of a ring with given inner and outer radius, centered around (0,0).

        **Tips:** *A ring with increasing inner radius, starting from 0, can look like an outfading point.*

        Parameters:
            out_r (int or float): The outer radius for the ring.
            in_r (int or float): The inner radius for the ring.

        Returns:
            A shape object representing a ring.
        """
        try:
            out_r2, in_r2 = out_r * 2, in_r * 2
            off = out_r - in_r
            off_in_r = off + in_r
            off_in_r2 = off + in_r2
        except TypeError:
            raise TypeError("Number(s) expected")

        if in_r >= out_r:
            raise ValueError(
                "Valid number expected. Inner radius must be less than outer radius"
            )

        f = Shape.format_value
        return Shape(
            "m 0 %s "
            "b 0 %s 0 0 %s 0 "
            "%s 0 %s 0 %s %s "
            "%s %s %s %s %s %s "
            "%s %s 0 %s 0 %s "
            "m %s %s "
            "b %s %s %s %s %s %s "
            "%s %s %s %s %s %s "
            "%s %s %s %s %s %s "
            "%s %s %s %s %s %s"
            % (
                f(out_r),  # outer move
                f(out_r),
                f(out_r),  # outer curve 1
                f(out_r),
                f(out_r2),
                f(out_r2),
                f(out_r),  # outer curve 2
                f(out_r2),
                f(out_r),
                f(out_r2),
                f(out_r2),
                f(out_r),
                f(out_r2),  # outer curve 3
                f(out_r),
                f(out_r2),
                f(out_r2),
                f(out_r),  # outer curve 4
                f(off),
                f(off_in_r),  # inner move
                f(off),
                f(off_in_r),
                f(off),
                f(off_in_r2),
                f(off_in_r),
                f(off_in_r2),  # inner curve 1
                f(off_in_r),
                f(off_in_r2),
                f(off_in_r2),
                f(off_in_r2),
                f(off_in_r2),
                f(off_in_r),  # inner curve 2
                f(off_in_r2),
                f(off_in_r),
                f(off_in_r2),
                f(off),
                f(off_in_r),
                f(off),  # inner curve 3
                f(off_in_r),
                f(off),
                f(off),
                f(off),
                f(off),
                f(off_in_r),  # inner curve 4
            )
        )

    @staticmethod
    def ellipse(w: float, h: float) -> Shape:
        """Returns a shape object of an ellipse with given width and height, centered around (0,0).

        **Tips:** *You could use that to create rounded stribes or arcs in combination with blurring for light effects.*

        Parameters:
            w (int or float): The width for the ellipse.
            h (int or float): The height for the ellipse.

        Returns:
            A shape object representing an ellipse.
        """
        try:
            w2, h2 = w / 2, h / 2
        except TypeError:
            raise TypeError("Number(s) expected")

        f = Shape.format_value

        return Shape(
            "m 0 %s "
            "b 0 %s 0 0 %s 0 "
            "%s 0 %s 0 %s %s "
            "%s %s %s %s %s %s "
            "%s %s 0 %s 0 %s"
            % (
                f(h2),  # move
                f(h2),
                f(w2),  # curve 1
                f(w2),
                f(w),
                f(w),
                f(h2),  # curve 2
                f(w),
                f(h2),
                f(w),
                f(h),
                f(w2),
                f(h),  # curve 3
                f(w2),
                f(h),
                f(h),
                f(h2),  # curve 4
            )
        )

    @staticmethod
    def heart(size: float, offset: float = 0) -> Shape:
        """Returns a shape object of a heart object with given size (width&height) and vertical offset of center point, centered around (0,0).

        **Tips:** *An offset=size*(2/3) results in a splitted heart.*

        Parameters:
            size (int or float): The width&height for the heart.
            offset (int or float): The vertical offset of center point.

        Returns:
            A shape object representing an heart.
        """
        try:
            mult = size / 30
        except TypeError:
            raise TypeError("Size parameter must be a number")
        # Build shape from template
        shape = Shape(
            "m 15 30 b 27 22 30 18 30 14 30 8 22 0 15 10 8 0 0 8 0 14 0 18 3 22 15 30"
        ).map(lambda x, y: (x * mult, y * mult))

        # Shift mid point of heart vertically
        count = 0

        def shift_mid_point(x, y):
            nonlocal count
            count += 1

            if count == 7:
                try:
                    return x, y + offset
                except TypeError:
                    raise TypeError("Offset parameter must be a number")
            return x, y

        # Return result
        return shape.map(shift_mid_point)

    @staticmethod
    def __glance_or_star(
        edges: int, inner_size: float, outer_size: float, g_or_s: str
    ) -> Shape:
        """
        General function to create a shape object representing star or glance.
        """
        # Alias for utility functions
        f = Shape.format_value

        def rotate_on_axis_z(point, theta):
            # Internal function to rotate a point around z axis by a given angle.
            theta = math.radians(theta)
            return Quaternion(axis=[0, 0, 1], angle=theta).rotate(point)

        # Building shape
        shape = ["m 0 %s %s" % (-outer_size, g_or_s)]
        inner_p, outer_p = 0, 0

        for i in range(1, edges + 1):
            # Inner edge
            inner_p = rotate_on_axis_z([0, -inner_size, 0], ((i - 0.5) / edges) * 360)
            # Outer edge
            outer_p = rotate_on_axis_z([0, -outer_size, 0], (i / edges) * 360)
            # Add curve / line
            if g_or_s == "l":
                shape.append(
                    "%s %s %s %s"
                    % (f(inner_p[0]), f(inner_p[1]), f(outer_p[0]), f(outer_p[1]))
                )
            else:
                shape.append(
                    "%s %s %s %s %s %s"
                    % (
                        f(inner_p[0]),
                        f(inner_p[1]),
                        f(inner_p[0]),
                        f(inner_p[1]),
                        f(outer_p[0]),
                        f(outer_p[1]),
                    )
                )

        shape = Shape(" ".join(shape))

        # Return result centered
        return shape.move()

    @staticmethod
    def star(edges: int, inner_size: float, outer_size: float) -> Shape:
        """Returns a shape object of a star object with given number of outer edges and sizes, centered around (0,0).

        **Tips:** *Different numbers of edges and edge distances allow individual n-angles.*

        Parameters:
            edges (int): The number of edges of the star.
            inner_size (int or float): The inner edges distance from center.
            outer_size (int or float): The outer edges distance from center.

        Returns:
            A shape object as a string representing a star.
        """
        return Shape.__glance_or_star(edges, inner_size, outer_size, "l")

    @staticmethod
    def glance(edges: int, inner_size: float, outer_size: float) -> Shape:
        """Returns a shape object of a glance object with given number of outer edges and sizes, centered around (0,0).

        **Tips:** *Glance is similar to Star, but with curves instead of inner edges between the outer edges.*

        Parameters:
            edges (int): The number of edges of the star.
            inner_size (int or float): The inner edges distance from center.
            outer_size (int or float): The control points for bezier curves between edges distance from center.

        Returns:
            A shape object as a string representing a glance.
        """
        return Shape.__glance_or_star(edges, inner_size, outer_size, "b")

    @staticmethod
    def rectangle(w: float = 1.0, h: float = 1.0) -> Shape:
        """Returns a shape object of a rectangle with given width and height, centered around (0,0).

        **Tips:** *A rectangle with width=1 and height=1 is a pixel.*

        Parameters:
            w (int or float): The width for the rectangle.
            h (int or float): The height for the rectangle.

        Returns:
            A shape object representing an rectangle.
        """
        try:
            f = Shape.format_value
            return Shape("m 0 0 l %s 0 %s %s 0 %s 0 0" % (f(w), f(w), f(h), f(h)))
        except TypeError:
            raise TypeError("Number(s) expected")

    @staticmethod
    def triangle(size: float) -> Shape:
        """Returns a shape object of an equilateral triangle with given side length, centered around (0,0).

        Parameters:
            size (int or float): The side length for the triangle.

        Returns:
            A shape object representing an triangle.
        """
        try:
            h = math.sqrt(3) * size / 2
            base = -h / 6
        except TypeError:
            raise TypeError("Number expected")

        f = Shape.format_value
        return Shape(
            "m %s %s l %s %s 0 %s %s %s"
            % (
                f(size / 2),
                f(base),
                f(size),
                f(base + h),
                f(base + h),
                f(size / 2),
                f(base),
            )
        )



# -*- coding: utf-8 -*-
# PyonFX: An easy way to create KFX (Karaoke Effects) and complex typesetting using the ASS format (Advanced Substation Alpha).
# Copyright (C) 2019 Antonio Strippoli (CoffeeStraw/YellowFlash)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyonFX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
"""
This module contains the Font class definition, which has some functions
to help getting informations from a specific font
"""
from __future__ import annotations
import sys
from typing import Tuple, TYPE_CHECKING

from .shape import Shape

if sys.platform == "win32":
    import win32gui  # pylint: disable=import-error
    import win32ui  # pylint: disable=import-error
    import win32con  # pylint: disable=import-error
elif sys.platform in ["linux", "darwin"] and not "sphinx" in sys.modules:
    import cairo  # pylint: disable=import-error
    import gi  # pylint: disable=import-error

    gi.require_version("Pango", "1.0")
    gi.require_version("PangoCairo", "1.0")

    from gi.repository import Pango, PangoCairo  # pylint: disable=import-error
    import html

if TYPE_CHECKING:
    from .ass_core import Style

# CONFIGURATION
FONT_PRECISION = 64  # Font scale for better precision output from native font system
LIBASS_FONTHACK = True  # Scale font data to fontsize? (no effect on windows)
PANGO_SCALE = 1024  # The PANGO_SCALE macro represents the scale between dimensions used for Pango distances and device units.


class Font:
    """
    Font class definition
    """

    def __init__(self, style: Style):
        self.family = style.fontname
        self.bold = style.bold
        self.italic = style.italic
        self.underline = style.underline
        self.strikeout = style.strikeout
        self.size = style.fontsize
        self.xscale = style.scale_x / 100
        self.yscale = style.scale_y / 100
        self.hspace = style.spacing
        self.upscale = FONT_PRECISION
        self.downscale = 1 / FONT_PRECISION

        if sys.platform == "win32":
            # Create device context
            self.dc = win32gui.CreateCompatibleDC(None)
            # Set context coordinates mapping mode
            win32gui.SetMapMode(self.dc, win32con.MM_TEXT)
            # Set context backgrounds to transparent
            win32gui.SetBkMode(self.dc, win32con.TRANSPARENT)
            # Create font handle
            font_spec = {
                "height": int(self.size * self.upscale),
                "width": 0,
                "escapement": 0,
                "orientation": 0,
                "weight": win32con.FW_BOLD if self.bold else win32con.FW_NORMAL,
                "italic": int(self.italic),
                "underline": int(self.underline),
                "strike out": int(self.strikeout),
                "charset": win32con.DEFAULT_CHARSET,
                "out precision": win32con.OUT_TT_PRECIS,
                "clip precision": win32con.CLIP_DEFAULT_PRECIS,
                "quality": win32con.ANTIALIASED_QUALITY,
                "pitch and family": win32con.DEFAULT_PITCH + win32con.FF_DONTCARE,
                "name": self.family,
            }
            self.pycfont = win32ui.CreateFont(font_spec)
            win32gui.SelectObject(self.dc, self.pycfont.GetSafeHandle())
            # Calculate metrics
            self.metrics = win32gui.GetTextMetrics(self.dc)
        elif sys.platform == "linux" or sys.platform == "darwin":
            surface = cairo.ImageSurface(cairo.Format.A8, 1, 1)

            self.context = cairo.Context(surface)
            self.layout = PangoCairo.create_layout(self.context)

            font_description = Pango.FontDescription()
            font_description.set_family(self.family)
            font_description.set_absolute_size(self.size * self.upscale * PANGO_SCALE)
            font_description.set_weight(
                Pango.Weight.BOLD if self.bold else Pango.Weight.NORMAL
            )
            font_description.set_style(
                Pango.Style.ITALIC if self.italic else Pango.Style.NORMAL
            )

            self.layout.set_font_description(font_description)
            self.metrics = Pango.Context.get_metrics(
                self.layout.get_context(), self.layout.get_font_description()
            )

            if LIBASS_FONTHACK:
                self.fonthack_scale = self.size / (
                    (self.metrics.get_ascent() + self.metrics.get_descent())
                    / PANGO_SCALE
                    * self.downscale
                )
            else:
                self.fonthack_scale = 1
        else:
            raise NotImplementedError

    def __del__(self):
        if sys.platform == "win32":
            win32gui.DeleteObject(self.pycfont.GetSafeHandle())
            win32gui.DeleteDC(self.dc)

    def get_metrics(self) -> Tuple[float, float, float, float]:
        if sys.platform == "win32":
            const = self.downscale * self.yscale
            return (
                # 'height': self.metrics['Height'] * const,
                self.metrics["Ascent"] * const,
                self.metrics["Descent"] * const,
                self.metrics["InternalLeading"] * const,
                self.metrics["ExternalLeading"] * const,
            )
        elif sys.platform == "linux" or sys.platform == "darwin":
            const = self.downscale * self.yscale * self.fonthack_scale / PANGO_SCALE
            return (
                # 'height': (self.metrics.get_ascent() + self.metrics.get_descent()) * const,
                self.metrics.get_ascent() * const,
                self.metrics.get_descent() * const,
                0.0,
                self.layout.get_spacing() * const,
            )
        else:
            raise NotImplementedError

    def get_text_extents(self, text: str) -> Tuple[float, float]:
        if sys.platform == "win32":
            cx, cy = win32gui.GetTextExtentPoint32(self.dc, text)

            return (
                (cx * self.downscale + self.hspace * (len(text) - 1)) * self.xscale,
                cy * self.downscale * self.yscale,
            )
        elif sys.platform == "linux" or sys.platform == "darwin":
            if not text:
                return 0.0, 0.0

            def get_rect(new_text):
                self.layout.set_markup(
                    f"<span "
                    f'strikethrough="{str(self.strikeout).lower()}" '
                    f'underline="{"single" if self.underline else "none"}"'
                    f">"
                    f"{html.escape(new_text)}"
                    f"</span>",
                    -1,
                )
                return self.layout.get_pixel_extents()[1]

            width = 0
            for char in text:
                width += get_rect(char).width

            return (
                (
                    width * self.downscale * self.fonthack_scale
                    + self.hspace * (len(text) - 1)
                )
                * self.xscale,
                get_rect(text).height
                * self.downscale
                * self.yscale
                * self.fonthack_scale,
            )
        else:
            raise NotImplementedError

    def text_to_shape(self, text: str) -> Shape:
        if sys.platform == "win32":
            # TODO: Calcultating distance between origins of character cells (just in case of spacing)

            # Add path to device context
            win32gui.BeginPath(self.dc)
            win32gui.ExtTextOut(self.dc, 0, 0, 0x0, None, text)
            win32gui.EndPath(self.dc)
            # Getting Path produced by Microsoft API
            points, type_points = win32gui.GetPath(self.dc)

            # Checking for errors
            if len(points) == 0 or len(points) != len(type_points):
                raise RuntimeError(
                    "This should never happen: function win32gui.GetPath has returned something unexpected.\nPlease report this to the developer"
                )

            # Defining variables
            shape, last_type = [], None
            mult_x, mult_y = self.downscale * self.xscale, self.downscale * self.yscale

            # Convert points to shape
            i = 0
            while i < len(points):
                cur_point, cur_type = points[i], type_points[i]

                if cur_type == win32con.PT_MOVETO:
                    if last_type != win32con.PT_MOVETO:
                        # Avoid repetition of command tags
                        shape.append("m")
                        last_type = cur_type
                    shape.extend(
                        [
                            Shape.format_value(cur_point[0] * mult_x),
                            Shape.format_value(cur_point[1] * mult_y),
                        ]
                    )
                    i += 1
                elif cur_type == win32con.PT_LINETO or cur_type == (
                    win32con.PT_LINETO | win32con.PT_CLOSEFIGURE
                ):
                    if last_type != win32con.PT_LINETO:
                        # Avoid repetition of command tags
                        shape.append("l")
                        last_type = cur_type
                    shape.extend(
                        [
                            Shape.format_value(cur_point[0] * mult_x),
                            Shape.format_value(cur_point[1] * mult_y),
                        ]
                    )
                    i += 1
                elif cur_type == win32con.PT_BEZIERTO or cur_type == (
                    win32con.PT_BEZIERTO | win32con.PT_CLOSEFIGURE
                ):
                    if last_type != win32con.PT_BEZIERTO:
                        # Avoid repetition of command tags
                        shape.append("b")
                        last_type = cur_type
                    shape.extend(
                        [
                            Shape.format_value(cur_point[0] * mult_x),
                            Shape.format_value(cur_point[1] * mult_y),
                            Shape.format_value(points[i + 1][0] * mult_x),
                            Shape.format_value(points[i + 1][1] * mult_y),
                            Shape.format_value(points[i + 2][0] * mult_x),
                            Shape.format_value(points[i + 2][1] * mult_y),
                        ]
                    )
                    i += 3
                else:  # If there is an invalid type -> skip, for safeness
                    i += 1

            # Clear device context path
            win32gui.AbortPath(self.dc)

            return Shape(" ".join(shape))
        elif sys.platform == "linux" or sys.platform == "darwin":
            # Defining variables
            shape, last_type = [], None

            def shape_from_text(new_text, x_add):
                nonlocal shape, last_type

                self.layout.set_markup(
                    f"<span "
                    f'strikethrough="{str(self.strikeout).lower()}" '
                    f'underline="{"single" if self.underline else "none"}"'
                    f">"
                    f"{html.escape(new_text)}"
                    f"</span>",
                    -1,
                )

                self.context.save()
                self.context.scale(
                    self.downscale * self.xscale * self.fonthack_scale,
                    self.downscale * self.yscale * self.fonthack_scale,
                )
                PangoCairo.layout_path(self.context, self.layout)
                self.context.restore()
                path = self.context.copy_path()

                # Convert points to shape
                for current_entry in path:
                    current_type = current_entry[0]
                    current_path = current_entry[1]

                    if current_type == 0:  # MOVE_TO
                        if last_type != current_type:
                            # Avoid repetition of command tags
                            shape.append("m")
                            last_type = current_type
                        shape.extend(
                            [
                                Shape.format_value(current_path[0] + x_add),
                                Shape.format_value(current_path[1]),
                            ]
                        )
                    elif current_type == 1:  # LINE_TO
                        if last_type != current_type:
                            # Avoid repetition of command tags
                            shape.append("l")
                            last_type = current_type
                        shape.extend(
                            [
                                Shape.format_value(current_path[0] + x_add),
                                Shape.format_value(current_path[1]),
                            ]
                        )
                    elif current_type == 2:  # CURVE_TO
                        if last_type != current_type:
                            # Avoid repetition of command tags
                            shape.append("b")
                            last_type = current_type
                        shape.extend(
                            [
                                Shape.format_value(current_path[0] + x_add),
                                Shape.format_value(current_path[1]),
                                Shape.format_value(current_path[2] + x_add),
                                Shape.format_value(current_path[3]),
                                Shape.format_value(current_path[4] + x_add),
                                Shape.format_value(current_path[5]),
                            ]
                        )

                self.context.new_path()

            curr_width = 0

            for i, char in enumerate(text):
                shape_from_text(char, curr_width + self.hspace * self.xscale * i)
                curr_width += self.get_text_extents(char)[0]

            return Shape(" ".join(shape))
        else:
            raise NotImplementedError



import os
import re
import setuptools

here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with open(os.path.join(here, *parts), "r") as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = [\'\"](.+)[\'\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


def get_requirements():
    requirements = ["pyquaternion"]

    if os.environ.get("READTHEDOCS") != "True":
        requirements.extend(
            [
                'pywin32; sys_platform == "win32"',
                'pycairo; sys_platform == "linux" or sys_platform == "darwin"',
                'PyGObject; sys_platform == "linux" or sys_platform == "darwin"',
            ]
        )

    return requirements


setuptools.setup(
    name="pyonfx",
    url="https://github.com/CoffeeStraw/PyonFX/",
    project_urls={
        "Documentation": "http://pyonfx.rtfd.io/",
        "Source": "https://github.com/CoffeeStraw/PyonFX/",
        "Tracker": "https://github.com/CoffeeStraw/PyonFX/issues/",
    },
    author="Antonio Strippoli",
    author_email="clarantonio98@gmail.com",
    description="An easy way to create KFX (Karaoke Effects) and complex typesetting using the ASS format (Advanced Substation Alpha).",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    version=find_version("pyonfx", "__init__.py"),
    packages=["pyonfx"],
    python_requires=">=3.7",
    install_requires=get_requirements(),
    extras_require={
        "dev": [
            "black",
            "pytest",
            "pytest-check",
            "sphinx_panels",
            "sphinx_rtd_theme",
            "sphinxcontrib-napoleon",
        ]
    },
    keywords="typesetting ass subtitle aegisub karaoke kfx advanced-substation-alpha karaoke-effect",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
    ],
    license="GNU LGPL 3.0 or later",
)
